/*
   touchcal - a tool to calibrate Touch Screens (namely ones with serial 
   controller manufactured by EloGraphics and MicroTouch) for use with XFree86.

   Copyright (C) 1999, 2002, 2004 Christoph Baumann <cgb@debian.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   The author wishes to thank Peter Baumann <Peter.Baumann@dlr.de> for the
   Linux Serial Programming HOWTO which was very helpfull during writing
   this program.

   Elo init code based on code from xf86-input-elographics, copied with
   permission from Patrick Lecoanet.
   Copyright 1995, 1999 by Patrick Lecoanet, France. <lecoanet@cena.dgac.fr>

*/

#ifdef HAVE_CONFIG_H
#   include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#   include <stdlib.h>
#endif


/*curses/ncurses library*/
#ifdef HAVE_LIBNCURSES
#   include <curses.h>
#else
#   ifdef HAVE_LIBCURSES
#       include <curses/curses.h>
#       ifndef getmaxx
#           define getmaxx(arg_) (arg_)->maxx
#           define getmaxy(arg_) (arg_)->maxy
#       endif
#   endif
#endif

#include <stdlib.h>
#include <curses.h>




/*produce something like a crosshair using curses to position it*/
static void cross(int position)
{
	static char * touchtxt = "Please touch the marker";
	int cx, cy, tx, ty;

	cx = cy = tx = ty = 0;

	clear();

	switch (position) {
		case 0:
			cx = 0; cy = 0;
			tx = 6; ty = 1;
			break;
		case 1:
			cx = 0; cy = getmaxy(stdscr) - 3;
			tx = 6; ty = 1;
			break;
		case 2:
			cx = getmaxx(stdscr) - 3; cy = 0;
			tx = -30; ty = 1;
			break;
		case 3:
			cx = getmaxx(stdscr) - 3; cy = getmaxy(stdscr) - 3;
			tx = -30; ty = 1;
			break;			
		case 4:
			cx = (getmaxx(stdscr)/2) - 3; cy = (getmaxy(stdscr)/2) - 3;
			tx = -30; ty = 1;
			break;			
		default:
			perror("invalid condition in cross()");
	}

	/*Paint cross*/
	move(cy, cx);
	printw(" | ");
	move(cy+1, cx);
	printw("-+-");
	move(cy+2, cx);
	printw(" | ");

	/*Print text*/
	move(cy+ty, cx+tx);
	printw(touchtxt);

	/*Center cursor in cross*/
	move(cy+1, cx+1);
	refresh();
}


static void show_result(int j, int * y, int * x)
{
	int i;
	clear();
	for (i = 0; i <= j; i++)
		mvprintw(10+i, 20, "Coord %d: y = %d, x = %d", i, y[i], x[i] );
	mvprintw(10+i+1, 23, "Hit ENTER to continue");
	getch();
}


/*extrapolish result and print it in XF86Config format*/
static void show_extrapolish_result(int * x, int * y, int sx, int sy)
{
	int xmin,xmax,ymin,ymax; /*counters and coordinates*/
	int screenx,screeny;
	FILE *out;

	screenx = sx - 3;
	screeny = sy - 3;

	/*extrapolish values*/
	xmax = x[2] + ((x[2] - x[0]) / screenx) * 1.5;
	xmin = x[0] - ((x[2] - x[0]) / screenx) * 1.5;
	ymax = y[0] + ((y[0] - y[1]) / screeny) * 1.5;
	ymin = y[1] - ((y[0] - y[1]) / screeny) * 1.5;

	/*print to stdout*/
	printf("   Option  \"MinX\" \"%d\"\n",xmin);
	printf("   Option  \"MaxX\" \"%d\"\n",xmax);
	printf("   Option  \"MinY\" \"%d\"\n",ymin);
	printf("   Option  \"MaxY\" \"%d\"\n",ymax);

	/*this output into a file*/
	out=fopen("/tmp/calib.tmp", "w");
	if (out == NULL)
		return;

	fprintf(out,"   Option  \"MinX\" \"%d\"\n",xmin);
	fprintf(out,"   Option  \"MaxX\" \"%d\"\n",xmax);
	fprintf(out,"   Option  \"MinY\" \"%d\"\n",ymin);
	fprintf(out,"   Option  \"MaxY\" \"%d\"\n",ymax);
	fclose(out);
}


/*calibration routine for an tablet device*/
static void tablet(char dev[])
{  
	unsigned char read, protocol[11];
	int i,j,x[4],y[4],xmin,xmax,ymin,ymax;
	int screenx,screeny;
	FILE *in;


	xmin = xmax = ymin = ymax = 0;

	/*in case we couldn't open the device*/
	in=fopen(dev,"rb");
	if (in==NULL) {
		fprintf(stderr,"\nCould not open device %s\n",dev); 
		return;
	}

	initscr();

	j = 0;

	/*read three coordinates*/
	while (j < 3) {
		/*display crosshair*/
		cross(j);

		/* drop strings not starting with UT */
		read = 'i';
		do {
			read = fgetc(in);
			if (read == 'U' && fgetc(in) == 'T')
				break;
		} while (1);

		/* read the rest of the word (9 bytes) - not counted, the next word
		 * starts with 'U' */
		i = 0;
		read = 'i';

		while (read != 'U') {
			read=fgetc(in);
			i++;
			if (read!='U') protocol[i]=read;
		}

		/*read X and Y coordinates*/      
		x[j] = protocol[3];
		x[j] = x[j] << 8;
		x[j] = x[j] | (protocol[2] & 0x00ff);     

		y[j] = protocol[5];
		y[j] = y[j] << 8;
		y[j] = y[j] | (protocol[4] & 0x00ff);    

		show_result(j, y, x);
		j++;
	} 
	screenx = getmaxx(stdscr);
	screeny = getmaxy(stdscr);

	endwin();
	fclose(in);

	/*finally show result*/
	show_extrapolish_result(x, y, screenx, screeny);

	return; 
}


int main(int anz, char *arg[])
{
	/*if not enough parameters are given*/
	if (anz < 2) {
		fprintf(stderr,"\nUsage: calib /dev/input/eventx [x=0..n]\n"); 
		return 1;
	}

	/*if the first parameter is an e*/
	printf("Calibrating tablet device\n");
	tablet(arg[1]);

	return 0;
}

