/*
   touchcal - a tool to calibrate Touch Screens (namely ones with serial 
   controller manufactured by EloGraphics and MicroTouch) for use with XFree86.

   Copyright (C) 1999, 2002, 2004 Christoph Baumann <cgb@debian.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   The author wishes to thank Peter Baumann <Peter.Baumann@dlr.de> for the
   Linux Serial Programming HOWTO which was very helpfull during writing
   this program.

   Elo init code based on code from xf86-input-elographics, copied with
   permission from Patrick Lecoanet.
   Copyright 1995, 1999 by Patrick Lecoanet, France. <lecoanet@cena.dgac.fr>

*/


#include "waltop-tablet.h"


#ifdef HAVE_CONFIG_H
#   include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#   include <stdlib.h>
#endif


/*curses/ncurses library*/
#ifdef HAVE_LIBNCURSES
#   include <curses.h>
#else
#   ifdef HAVE_LIBCURSES
#       include <curses/curses.h>
#       ifndef getmaxx
#           define getmaxx(arg_) (arg_)->maxx
#           define getmaxy(arg_) (arg_)->maxy
#       endif
#   endif
#endif


/*****************************************************************************
** Defines
*****************************************************************************/

/* from linux/input.h */
#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define BIT(x)  (1UL<<((x)%BITS_PER_LONG))
#define LONG(x) ((x)/BITS_PER_LONG)
#define ISBITSET(x,y) ((x)[LONG(y)] & BIT(y))

/*****************************************************************************
** Structures
*****************************************************************************/

struct ABS_STATE
{
	int bValid;
	int nRow, nCol;
	int nValue, nMin, nMax;
};

struct KEY_STATE
{
	int bValid;
	int nRow, nCol;
	int nValue;
};

struct ABS_STATE gAbsState[WALTOPFIELD_MAX];
struct KEY_STATE gKeyState[WALTOPBUTTON_MAX];

/*int gYCross=-1;
int gXCross=-1;*/

/*****************************************************************************
** tablet function
*****************************************************************************/

static int InitTablet(WALTOPTABLET hTablet)
{
	int i, nCaps, nItem, nRow=0;
	WALTOPSTATE ranges = WALTOPSTATE_INIT;


	/* get event types supported, ranges, and immediate values */
	nCaps = WaltopGetCapabilities(hTablet);
	WaltopGetState(hTablet,&ranges);

	nItem = 0;
	for (i=0; i<31; ++i)
	{
		if (nCaps & (1<<i))
		{
			gAbsState[i].bValid = 1;
			gAbsState[i].nValue = ranges.values[i].nValue;
			gAbsState[i].nMin = ranges.values[i].nMin;
			gAbsState[i].nMax = ranges.values[i].nMax;
			gAbsState[i].nRow = nRow + nItem / 2;
			gAbsState[i].nCol = nItem % 2;
			++nItem;
		}
	}
	nRow += ((nItem + 1) / 2) + 1;

	/* get key event types */
	nItem = 0;
	for (i=0; i<WALTOPBUTTON_MAX; ++i)
	{
		gKeyState[i].bValid = 1;
		gKeyState[i].nRow = nRow + nItem / 4;
		gKeyState[i].nCol = nItem % 4;
		++nItem;
	}
	nRow += ((nItem + 3) / 4) + 1;
	return 0;
}

/* 
 * [in] fetch: 1=fetch tablet, 0=flush tablet
 */
static int FetchTablet(WALTOPTABLET hTablet, int fetch)
{
	int i, len;
	unsigned char data[64];
	WALTOPSTATE state = WALTOPSTATE_INIT;	


	gKeyState[WALTOPBUTTON_TOUCH].nValue = 0;
	
	while(1){
		len = WaltopReadRaw(hTablet, data, sizeof(data));
		if(len<0){
			sleep(1);
			continue;
		}
			
		if((i = WaltopParseData(hTablet, data, len, &state))){
			/*fprintf(stdout, "ParseData - %d\n", i);*/
			continue;
		}

		for (i=WALTOPFIELD_TOOLTYPE; i<WALTOPFIELD_MAX; ++i)
		{
			if (state.uValid & BIT(i))
			{
				gAbsState[i].nValue = state.values[i].nValue;					
			}
		}
		
		for(i=WALTOPBUTTON_LEFT; i<WALTOPBUTTON_MAX; ++i){
			gKeyState[i].nValue = state.values[WALTOPFIELD_BUTTONS].nValue & (1<<i) ? 1 : 0;
		}
		
		if(gKeyState[WALTOPBUTTON_TOUCH].nValue == fetch){
			break;
		}
	}
}

/*****************************************************************************
** calibration function
*****************************************************************************/

/*produce something like a crosshair using curses to position it*/
static void cross(int position)
{
	static char * touchtxt = "Please touch the marker";
	int cx, cy, tx, ty;

	cx = cy = tx = ty = 0;

	clear();

	switch (position) {
		case 0:
			cx = 0; cy = 0;
			tx = 6; ty = 1;
			break;
		case 1:
			cx = 0; cy = getmaxy(stdscr) - 3;
			tx = 6; ty = 1;
			break;
		case 2:
			cx = getmaxx(stdscr) - 3; cy = 0;
			tx = -30; ty = 1;
			break;
		case 3:
			cx = getmaxx(stdscr) - 3; cy = getmaxy(stdscr) - 3;
			tx = -30; ty = 1;
			break;			
		case 4:
			cx = (getmaxx(stdscr)/2) - 3; cy = (getmaxy(stdscr)/2) - 3;
			tx = -30; ty = 1;
			break;			
		default:
			perror("invalid condition in cross()");
	}

	/*Paint cross*/
	move(cy, cx);
	printw(" | ");
	move(cy+1, cx);
	printw("-+-");
	move(cy+2, cx);
	printw(" | ");

	/*Print text*/
	move(cy+ty, cx+tx);
	printw(touchtxt);

	/*Center cursor in cross*/
	move(cy+1, cx+1);
	/*getyx(stdscr, gYCross, gXCross);*/	
	refresh();
}


static void show_result(int j, int * y, int * x)
{
	int i;
	clear();
	for (i = 0; i <= j; i++)
		mvprintw(10+i, 20, "Coord %d: y = %d, x = %d", i, y[i], x[i] );
	mvprintw(10+i+1, 23, "Hit ENTER to continue");
	getch();
}


/*extrapolish result and print it in XF86Config format*/
static void show_extrapolish_result(int * x, int * y, int sx, int sy)
{
	int xmin,xmax,ymin,ymax; /*counters and coordinates*/
	int screenx,screeny;
	FILE *out;

	screenx = sx - 3;
	screeny = sy - 3;

	/*extrapolish values*/
	xmax = x[2] + ((x[2] - x[0]) / screenx) * 1.5;
	xmin = x[0] - ((x[2] - x[0]) / screenx) * 1.5;
	ymax = y[0] + ((y[0] - y[1]) / screeny) * 1.5;
	ymin = y[1] - ((y[0] - y[1]) / screeny) * 1.5;

	/*print to stdout*/
	printf("   Option  \"MinX\" \"%d\"\n",xmin);
	printf("   Option  \"MaxX\" \"%d\"\n",xmax);
	printf("   Option  \"MinY\" \"%d\"\n",ymin);
	printf("   Option  \"MaxY\" \"%d\"\n",ymax);

	/*this output into a file*/
	out=fopen("/tmp/calib.tmp", "w");
	if (out == NULL)
		return;

	fprintf(out,"   Option  \"MinX\" \"%d\"\n",xmin);
	fprintf(out,"   Option  \"MaxX\" \"%d\"\n",xmax);
	fprintf(out,"   Option  \"MinY\" \"%d\"\n",ymin);
	fprintf(out,"   Option  \"MaxY\" \"%d\"\n",ymax);
	fclose(out);
}


/*calibration routine for an tablet device*/
static void Tablet(char dev[])
{  
	int i,j,x[4],y[4],xmin,xmax,ymin,ymax;
	int xcross, ycross;
	int screenx,screeny;
	
	WALTOPTABLET hTablet = NULL;
	
	
	xmin = xmax = ymin = ymax = 0;

	/*open the device*/
	hTablet = WaltopOpenTablet(dev);
	if (!hTablet)
	{
		perror("WaltopOpenTablet");
		exit(1);
	}
	
	/* init device*/
	if(InitTablet(hTablet)){
		perror("InitTablet");
		WaltopCloseTablet(hTablet);
		exit(1);
	}

	initscr();

	j = 0;	

	/*read three coordinates*/
	while (j < 3) {
		/*display crosshair*/
		cross(j);

		/* fetch tablet */
		FetchTablet(hTablet, 1);

		if(gKeyState[WALTOPBUTTON_TOUCH].nValue == 1){
			x[j] = gAbsState[WALTOPFIELD_POSITION_X].nValue;
			y[j] = gAbsState[WALTOPFIELD_POSITION_Y].nValue;
			beep();
		}

		/*flush buffer*/
		i=0;
		while(i<20){
			FetchTablet(hTablet, 0);
			++i;
		}
		
		show_result(j, y, x);		
		j++;	
	} 
	screenx = getmaxx(stdscr);
	screeny = getmaxy(stdscr);

	endwin();
	WaltopCloseTablet(hTablet);

	/*finally show result*/
	show_extrapolish_result(x, y, screenx, screeny);

	return; 
}


int main(int anz, char *arg[])
{
	/*if not enough parameters are given*/
	if (anz < 2) {
		fprintf(stderr,"\nUsage: calib /dev/input/eventx [x=0..n]\n"); 
		return 1;
	}

	/*if the first parameter is an e*/
	printf("Calibrating tablet device\n");
	Tablet(arg[1]);

	return 0;
}

