/*
 * based on linuxwacom-0.8.0-3
 */


#include "waltop-usb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <assert.h>


#include <linux/input.h>


/*****************************************************************************
** Defines
*****************************************************************************/

/* from linux/input.h */
#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define BIT(x)  (1UL<<((x)%BITS_PER_LONG))
#define LONG(x) ((x)/BITS_PER_LONG)
#define ISBITSET(x,y) ((x)[LONG(y)] & BIT(y))


#define MAX_CHANNELS 2
#define MAX_USB_EVENTS 32

/*****************************************************************************
** USB Tablet object
*****************************************************************************/

typedef struct _USBTABLET USBTABLET;

struct _USBTABLET
{
	WALTOPTABLET_PRIV tablet;
	int fd;
	WALTOPSTATE state[MAX_CHANNELS];
	int nToolProx[MAX_CHANNELS];
	int nChannel;
	int nEventCnt;
	struct input_event events[MAX_USB_EVENTS];
};


/*****************************************************************************
** Static Prototypes
*****************************************************************************/

static void USBClose(WALTOPTABLET_PRIV* pTablet);

static int USBGetCaps(WALTOPTABLET_PRIV* pTablet);

static int USBGetState(WALTOPTABLET_PRIV* pTablet, WALTOPSTATE* pState);

static int USBGetFD(WALTOPTABLET_PRIV* pTablet);

static void USBSetFD(WALTOPTABLET_PRIV* pTablet, int fd);

static int USBReadRaw(WALTOPTABLET_PRIV* pTablet, unsigned char* puchData,
		unsigned int uSize);

static int USBParseData(WALTOPTABLET_PRIV* pTablet,
		const unsigned char* puchData, unsigned int uLength,
		WALTOPSTATE* pState);

static int USBIdentifyModel(USBTABLET* pUSB);


/*****************************************************************************
** Public Functions
*****************************************************************************/

WALTOPTABLET OpenUSBTablet(int fd)
{
	USBTABLET* pUSB = NULL;

	/* Allocate tablet */
	pUSB = (USBTABLET*)malloc(sizeof(USBTABLET));
	memset(pUSB,0,sizeof(*pUSB));
	pUSB->tablet.Close = USBClose;
	pUSB->tablet.GetCaps = USBGetCaps;
	pUSB->tablet.GetState = USBGetState;
	pUSB->tablet.GetFD = USBGetFD;
	pUSB->tablet.SetFD = USBSetFD;
	pUSB->tablet.ReadRaw = USBReadRaw;
	pUSB->tablet.ParseData = USBParseData;
	pUSB->fd = fd;

	/* Identify and initialize the model */
	if (USBIdentifyModel(pUSB))
	{
		fprintf(stderr, "Failed to identify model: %s",strerror(errno));
		free(pUSB);
		return NULL;
	}

	return (WALTOPTABLET)pUSB;
}

/*****************************************************************************
** Tablet functions
*****************************************************************************/

static void USBClose(WALTOPTABLET_PRIV* pTablet)
{
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	close(pUSB->fd);
	free(pUSB);
}

static int USBGetRange(USBTABLET* pUSB, unsigned long* pBits, int nAbsField,
		unsigned int uField)
{
	int nAbs[5];

	if  (!ISBITSET(pBits,nAbsField))
		return 0;
	if (ioctl(pUSB->fd, EVIOCGABS(nAbsField), nAbs) != 0)
		return 1;

	pUSB->state[0].values[uField].nMin = nAbs[1];
	pUSB->state[0].values[uField].nMax = nAbs[2];
	pUSB->state[0].uValid |= BIT(uField);
	return 0;
}

static int USBIdentifyModel(USBTABLET* pUSB)
{
	int nCnt, chcnt;
	short sID[4];
	unsigned long evbits[NBITS(EV_MAX)];
	unsigned long absbits[NBITS(ABS_MAX)];
	unsigned long relbits[NBITS(REL_MAX)];
	unsigned long keybits[NBITS(KEY_MAX)];

	/* Get device name and id */
	if (ioctl(pUSB->fd,EVIOCGID,sID) < 0)
		return 1;

	/* initialize state structure */
	pUSB->state[0].uValueCnt = WALTOPFIELD_MAX;

	/* Get event types supported */
	nCnt = ioctl(pUSB->fd,EVIOCGBIT(0 /*EV*/,sizeof(evbits)),evbits);
	if (nCnt < 0)
	{
		fprintf(stderr, "Failed to CGBIT ev: %s",strerror(errno));
		return 1;
	}
	assert(nCnt == sizeof(evbits));

	/* absolute events */
	if (ISBITSET(evbits,EV_ABS))
	{
		nCnt = ioctl(pUSB->fd,EVIOCGBIT(EV_ABS,sizeof(absbits)),absbits);
		if (nCnt < 0)
		{
			fprintf(stderr, "Failed to CGBIT abs: %s",strerror(errno));
			return 1;
		}

		/* the following line has problem on Debian systems
		assert(nCnt == sizeof(absbits));
		*/

		if (USBGetRange(pUSB,absbits,ABS_X,WALTOPFIELD_POSITION_X) ||
			USBGetRange(pUSB,absbits,ABS_Y,WALTOPFIELD_POSITION_Y) ||			
			USBGetRange(pUSB,absbits,ABS_PRESSURE,WALTOPFIELD_PRESSURE))
			return 1;
	}

	/* relative events */
	if (ISBITSET(evbits,EV_REL))
	{
		nCnt = ioctl(pUSB->fd,EVIOCGBIT(EV_REL,sizeof(relbits)),relbits);
		if (nCnt < 0)
		{
			fprintf(stderr, "Failed to CGBIT rel: %s",strerror(errno));
			return 1;
		}
		assert(nCnt == sizeof(relbits));
	}

	/* key events */
	if (ISBITSET(evbits,EV_KEY))
	{
		nCnt = ioctl(pUSB->fd,EVIOCGBIT(EV_KEY,sizeof(keybits)),keybits);
		if (nCnt < 0)
		{
			fprintf(stderr, "Failed to CGBIT key: %s",strerror(errno));
			return 1;
		}
		assert(nCnt == sizeof(keybits));

		/* button events */
		if (ISBITSET(keybits,BTN_LEFT) ||
				ISBITSET(keybits,BTN_RIGHT) ||
				ISBITSET(keybits,BTN_MIDDLE) ||
				ISBITSET(keybits,BTN_SIDE) ||
				ISBITSET(keybits,BTN_EXTRA))
			pUSB->state[0].uValid |= BIT(WALTOPFIELD_BUTTONS);

		/* tool events */
		if (ISBITSET(keybits,BTN_TOOL_PEN))
			pUSB->state[0].uValid |= BIT(WALTOPFIELD_PROXIMITY) |
					BIT(WALTOPFIELD_TOOLTYPE);
	}


	/* Initialize all channels with the same values */
	for(chcnt=1; chcnt<MAX_CHANNELS; chcnt++)
		pUSB->state[chcnt] = pUSB->state[0];

	return 0;
}


static int USBGetCaps(WALTOPTABLET_PRIV* pTablet)
{
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	return pUSB->state[0].uValid;
}


static int USBGetState(WALTOPTABLET_PRIV* pTablet, WALTOPSTATE* pState)
{
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	return WaltopCopyState(pState,&pUSB->state[pUSB->nChannel]);
}


static int USBGetFD(WALTOPTABLET_PRIV* pTablet)
{
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	return pUSB->fd;
}

static void USBSetFD(WALTOPTABLET_PRIV* pTablet, int fd)
{
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	pUSB->fd = fd;
}

static int USBReadRaw(WALTOPTABLET_PRIV* pTablet, unsigned char* puchData,
		unsigned int uSize)
{
	int nXfer;
	unsigned int uCnt, uPacketLength;
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	uPacketLength = sizeof(struct input_event);

	/* check size of buffer */
	if (uSize < uPacketLength) { errno=EINVAL; return 0; }
	
	for (uCnt=0; uCnt<uPacketLength; uCnt+=nXfer)
	{
		nXfer = read(pUSB->fd,puchData+uCnt,uPacketLength-uCnt);
		if (nXfer <= 0) return nXfer;
	}

	return (signed)uCnt;
}


static int USBParseMSC(USBTABLET* pUSB, struct input_event* pEv)
{
	return 0;
}


static int USBParseKEY(USBTABLET* pUSB, struct input_event* pEv)
{
	int i, button=-1, tool=0;
	switch (pEv->code)
	{
		case BTN_LEFT: button = WALTOPBUTTON_LEFT; break;
		case BTN_RIGHT: button = WALTOPBUTTON_RIGHT; break;
		case BTN_MIDDLE: button = WALTOPBUTTON_MIDDLE; break;
		case BTN_SIDE: button = WALTOPBUTTON_SIDE; break;
		case BTN_EXTRA: button = WALTOPBUTTON_EXTRA; break;
		case BTN_TOUCH: button = WALTOPBUTTON_TOUCH; break;
		case BTN_STYLUS: button = WALTOPBUTTON_STYLUS; break;
		case BTN_TOOL_PEN: tool = WALTOPTOOLTYPE_PEN; break;
	}

	/* if button was specified */
	if (button != -1)
	{
		/* button state change */
		if (pEv->value)
			pUSB->state[pUSB->nChannel].values[WALTOPFIELD_BUTTONS].nValue |= BIT(button);
		else
			pUSB->state[pUSB->nChannel].values[WALTOPFIELD_BUTTONS].nValue &= ~BIT(button);
	}

	/* if a tool was specified */
	if (tool)
	{
		/* coming into proximity */
		if (pEv->value)
		{
			/* no prior tool in proximity */
			if (!(pUSB->nToolProx[pUSB->nChannel] & BIT(tool)))
			{
				pUSB->state[pUSB->nChannel].values[WALTOPFIELD_PROXIMITY].nValue = 1;
			}

			/* remember tool in prox */
			pUSB->nToolProx[pUSB->nChannel] |= BIT(tool);
			pUSB->state[pUSB->nChannel].values[WALTOPFIELD_TOOLTYPE].nValue = tool;
		}

		/* otherwise, going out of proximity */
		else
		{
			/* forget tool in prox */
			if (pUSB->nToolProx[pUSB->nChannel] & BIT(tool))
			{
				pUSB->nToolProx[pUSB->nChannel] &= ~BIT(tool);
			}

			/* single input */
			/*if (!(pUSB->state[pUSB->nChannel].uValid & BIT(WALTOPFIELD_SERIAL)))*/
				pUSB->nToolProx[pUSB->nChannel] = 0;

			pUSB->state[pUSB->nChannel].values[WALTOPFIELD_PROXIMITY].nValue = 0;
			pUSB->state[pUSB->nChannel].values[WALTOPFIELD_TOOLTYPE].nValue = 0;

			/* nobody left? out of proximity */
			if (!pUSB->nToolProx[pUSB->nChannel])
				memset(pUSB->state[pUSB->nChannel].values, 0,
						pUSB->state[pUSB->nChannel].uValueCnt * sizeof(WALTOPVALUE));
			/* otherwise, switch to next tool */
			else
			{
				for (i=WALTOPTOOLTYPE_PEN; i<WALTOPTOOLTYPE_MAX; ++i)
				{
					if (pUSB->nToolProx[pUSB->nChannel] & BIT(i))
					{
						pUSB->state[pUSB->nChannel].values[WALTOPFIELD_TOOLTYPE].nValue = i;
					}
				}
			}
		} /* out of prox */
	} /* end if tool */

	return 0;
}


static int USBParseABS(USBTABLET* pUSB, struct input_event* pEv)
{
	int field = 0;
	switch (pEv->code)
	{
		case ABS_X: field = WALTOPFIELD_POSITION_X; break;
		case ABS_Y: field = WALTOPFIELD_POSITION_Y; break;
		case ABS_PRESSURE: field = WALTOPFIELD_PRESSURE; break;
	}

	if (field)
		pUSB->state[pUSB->nChannel].values[field].nValue = pEv->value;

	return 0;
}


static int USBParseData(WALTOPTABLET_PRIV* pTablet,
		const unsigned char* puchData, unsigned int uLength,
		WALTOPSTATE* pState)
{
	int evcnt, chcnt;
	USBTABLET* pUSB = (USBTABLET*)pTablet;
	struct input_event* pEv = (struct input_event*)puchData;
	if (uLength != sizeof(struct input_event)) return 1;

	/* store event until we receive a MSC_SERIAL/SYN_REPORT
	 * so we can figure out with channel to use */
	if (pUSB->nEventCnt >= MAX_USB_EVENTS)
	{
		/* no more buffer space */
		pUSB->nEventCnt = 0;
		errno = ENOBUFS;
		return -1;
	}
	pUSB->events[pUSB->nEventCnt++] = *pEv;

	if ((pEv->type == EV_MSC) && (pEv->code == MSC_SERIAL))
	{
		/* store the serial for the tool for later use */
#ifdef EV_SYN
		/* Kernel 2.4 uses MSC_SERIAL as an end-of-report, 2.6
		 * don't so we're not done yet */
		return 0;
	}
	else if ((pEv->type == EV_SYN) && (pEv->code == SYN_REPORT))
	{
		/* Kernel 2.6 used SYN_REPORT as an end-of-record,
		 * fall through */
#endif
	}
	else
	{
		/* Not a MSC_SERIAL or SYN_REPORT, we're not done yet */
		return 0;
	}

	/* Select channel here based on the serial number, tablets
	 * with only one channel will always use channel 0, so no
	 * check.
	 */

	/* Channel found/allocated, lets process events */

	for (evcnt=0; evcnt<pUSB->nEventCnt; evcnt++)
	{
		pEv = pUSB->events + evcnt;
		/* dispatch event */
		switch (pEv->type)
		{
#ifdef EV_SYN
		case EV_SYN: /* kernel 2.6 */
#endif
		case EV_MSC: /* kernel 2.4 */
			     if (USBParseMSC(pUSB,pEv)) return pEv->type; break;
		case EV_KEY: if (USBParseKEY(pUSB,pEv)) return pEv->type; break;
		case EV_ABS: if (USBParseABS(pUSB,pEv)) return pEv->type; break;
		default: errno = EINVAL; return pEv->type;
		}
	}

	pUSB->nEventCnt = 0;
	return pState ? WaltopCopyState(pState,pUSB->state + pUSB->nChannel) : 0;
}


